/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * The game of hangman.
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <time.h>

#include <dpy.h>


#ifndef	WORDFILE
#define	WORDFILE	"/usr/share/dict/words"
#endif

#define	BUFSIZE		8192		/* buffer size for word storage */
#define	MINWORDCOUNT	100		/* minimum number of words in file */
#define	LISTALLOC	1024		/* allocation increment for word list */
#define	MINLEN		7		/* default minimum word length */
#define	MAXLEN		15		/* maximum word length */
#define	LOSTERRORS	8		/* number of errors to lose game */
#define	PICTCOL		42		/* column where picture starts */
#define	ESC		'\033'		/* command to quit game */

#define	AVAIL		0		/* letter is available */
#define	RIGHT		1		/* letter was rightly guessed */
#define	WRONG		2		/* letter was bad guess */

#define	TRUE		((BOOL) 1)	/* true value */
#define	FALSE		((BOOL) 0)	/* false value */


typedef	int BOOL;


static	const char * wordFile = WORDFILE; /* path of words file */
static	int minLen = MINLEN;		/* minimum length of words */
static	BOOL isCapitalsAllowed;		/* TRUE if capitals allowed */
static	char ** wordList;		/* the list of words */
static	int wordListSize;		/* the number of words in the list */


/*
 * The picture of the man.
 * Characters are in pairs where the first is the display character,
 * and the second is what move to display it.  However, backquotes are
 * used here to indicate backslashes in order to avoid quoting problems.
 */
static	const char * pict[] =
{
"      + - - - - - - - - - - - - - - - - - - - - - - +                     ",
"      |   + - - - - /   / - - - - - - - - - - | - - +                     ",
"      |   |       /   /                       |                           ",
"      |   |     /   /                   +1-1-1-1-1-1+1                    ",
"      |   |   /   /                     |1  o2  o2  |1                    ",
"      |   | /   /                       |1    .2    |1                    ",
"      |   /   /                         |1  /8=2`8  |1                    ",
"      |     /                           +1-1+1  +1-1+1                    ",
"      |   |                                 |1  |1                        ",
"      |   |                       -4-4+3-3-3+3  +3-3-3+3-5-5              ",
"      |   |                     /4    |3              |3    `5            ",
"      |   |                   /4    /4|3              |3`5    `5          ",
"      |   |                 /4    /4  |3              |3  `5    `5        ",
"      |   |               /4    /4    |3              |3    `5    `5      ",
"      |   |             |4    /4      +3-3-3-3-3-3-3-3+3      `5    |5    ",
"      |   |             +4-4/4        |6      |6      |7        `5-5+5    ",
"      |   |                         /6      /6  `7      `7                ",
"      |   |                       /6      /6      `7      `7              ",
"      |   |                     /6      /6          `7      `7            ",
"      |   |                   /6      /6              `7      `7          ",
"      |   |           /6-6-6+6      |6                  |7      +7-7-7`7  ",
"      |   |           +6-6-6-6-6-6-6+6                  +7-7-7-7-7-7-7-7+7",
"      |   |                                                               ",
"      |   |                                                               ",
"      |   |                                                               ",
"- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ",
NULL
};


/*
 * Local routines.
 */
static	void readWordList(void);
static	const char * pickWord(void);
static	int randomRange(int);
static	BOOL playGame(void);
static	void drawPict(int);
static	void usage(void);
static	void fileError(const char *);


int
main(int argc, const char ** argv)
{
	long curTime = (long) time(0);

	srandom((int) (curTime ^ (curTime >> 16)));

	argc--;
	argv++;

	while ((argc > 0) && (**argv == '-'))
	{
		const char * cp = &argv[0][1];

		while (*cp) switch (*cp++)
		{
			case 'c':
				/* allow capitals in words */
				isCapitalsAllowed = TRUE;
				break;

			default:
				usage();
		}

		argc--;
		argv++;
	}

	if (argc > 0)
	{
		wordFile = *argv++;
		argc--;
	}

	if (argc > 0)
		usage();

	readWordList();

	if (dpyinit(0, 0) < 0)
		return 1;

	while (playGame())
		;

	dpyclose();

	return 0;
}


/*
 * Play one game of hangman.
 * Returns TRUE if want to play another game.
 */
static BOOL
playGame(void)
{
	const char * rawWord;	/* word to guess */
	const char * word;	/* word to display */
	const char * cp;	/* character */
	int wordLength;		/* length of word */
	int charsLeft;		/* characters left to be guessed */
	int guesses;		/* number of guesses taken */
	int errorCount;		/* number of errors made */
	int ch;			/* current guess */
	int state;		/* state of current letter */
	int i;			/* index */
	char lowWord[MAXLEN];	/* lower cased word */
	char seeWord[MAXLEN];	/* word user sees */
	char letters[26];	/* table of letters */
	char buf[PICTCOL];	/* buffer */

	rawWord = pickWord();
	wordLength = strlen(rawWord);
	charsLeft = wordLength;
	guesses = 0;
	errorCount = 0;

	for (i = 0; i < wordLength; i++)
	{
		ch = rawWord[i];

		if ((ch >= 'A') && (ch <= 'Z'))
			ch += ('a' - 'A');

		lowWord[i] = ch;
		seeWord[i] = '-';
	}

	memset(letters, AVAIL, 26);

	dpywindow(0, -1, 0, -1);
	dpyclrwindow();

	for (;;)
	{
		drawPict(errorCount);

		dpywindow(0, -1, 0, PICTCOL - 1);
		dpymove(3, 0);
		dpyprintf("Guesses taken: %d\n", guesses);
		dpyprintf("Errors made:   %d\n", errorCount);
		dpymove(9, 2);

		for (ch = 'A'; ch <= 'Z'; ch++)
		{
			if (letters[ch - 'A'] != AVAIL)
				dpychar('.');
			else
				dpychar(ch);

			dpystr("  ");

			if (ch == 'M')
				dpymove(11, 2);
		}

		dpymove(15, 0);

		i = (PICTCOL - 4 - (wordLength * 2 - 1)) / 2;

		while (i-- > 0)
			dpychar(' ');

		word = ((errorCount >= LOSTERRORS) ? rawWord : seeWord);

		for (i = 0; i < wordLength; i++)
			dpyprintf("%c ", *word++);

		dpyclrwindow();

		if ((charsLeft <= 0) || (errorCount >= LOSTERRORS))
			break;

		dpymove(0, 0);
		dpystr("Guess: ");
		dpyupdate();

		ch = dpyinputchar();

		if ((ch < 0) || (ch == ESC))
			return FALSE;

		if (ch == ('L' - '@'))
		{
			dpyredraw();
			continue;
		}

		if ((ch >= 'A') && (ch <= 'Z'))
			ch += ('a' - 'A');

		if ((ch < 'a') || (ch > 'z'))
			continue;

		if (letters[ch - 'a'] != AVAIL)
			continue;

		state = WRONG;

		for (i = 0; i < wordLength; i++)
		{
			if (ch != lowWord[i])
				continue;

			seeWord[i] = rawWord[i];
			charsLeft--;
			state = RIGHT;
		}

		letters[ch - 'a'] = state;
		guesses++;

		if (state == WRONG)
			errorCount++;
	}

	dpywindow(0, 0, 0, PICTCOL - 1);

	/*
	 * Show the game result and read whether or not to play again.
	 */
	if (errorCount >= LOSTERRORS)
		cp = "Oh no, you lost.  Play again? ";
	else if (errorCount == 0)
		cp = "Great, you guessed perfectly!  Play again? ";
	else
		cp = "Hurray, you got it.  Play again? ";

	dpyread(cp, 0, buf, sizeof(buf), 0);

	cp = buf;

	while (*cp == ' ')
		cp++;

	return ((*cp != 'n') && (*cp != 'N'));
}


/*
 * Draw the picture of the man corresponding to the specified
 * number of errors.
 */
static void
drawPict(int errorCount)
{
	const char ** line;
	int level;

	dpywindow(0, -1, PICTCOL, -1);

	for (line = pict; *line; line++)
	{
		const char * cp = *line;

		while (*cp)
		{
			int ch = *cp++;

			if (ch == '`')
				ch = '\\';

			level = *cp++;

			if ((level != ' ') && (level > (errorCount + '0')))
				ch = ' ';

			dpychar(ch);
		}

		dpychar('\n');
	}
}


/*
 * Read the dictionary into memory and build a word list to point to
 * the valid words from it.  Words cannot be capitalized unless the -c
 * option was given, and must be between minLen and MAXLEN chars in length.
 * Doesn't return on errors.
 */
static void
readWordList(void)
{
	char * buffer = 0;
	int bufferLeft = 0;
	int wordListAlloc = 0;

	FILE * fp = fopen(wordFile, "r");

	char line[256];

	if (fp == 0)
		fileError("Open failed");

	/*
	 * Read each line of the dictionary file (which is a possible word).
	 * Append all of the suitable words found into the word list.
	 */
	wordList = 0;
	wordListSize = 0;

	while (fgets(line, sizeof(line), fp) != 0)
	{
		int len = strlen(line) - 1;
		BOOL isBadWord = FALSE;
		BOOL isCapWord = FALSE;
		const char * cp = line;

		/*
		 * If the line doesn't end in a newline or the length
		 * of the word isn't in our range then skip it.
		 * Here the length doesn't include the terminating newline.
		 * Then null terminate the word and count the null.
		 */
		if (line[len] != '\n')
			continue;

		if ((len < minLen) || (len > MAXLEN))
			continue;

		line[len++] = '\0';

		/*
		 * Check the word's characters to see if it is acceptable.
		 * This means that it only contains alphabetic characters
		 * and possibly only contains lower case.
		 */
		while (*cp)
		{
			if ((*cp >= 'A') && (*cp <= 'Z'))
				isCapWord = TRUE;
			else if ((*cp < 'a') || (*cp > 'z'))
				isBadWord = TRUE;

			cp++;
		}

		if (isBadWord || (isCapWord && !isCapitalsAllowed))
			continue;

		/*
		 * This is a good word so we can save it away.
		 * Allocate a new storage buffer if necessary.
		 * Note: Any previous buffer is forgotten but will still
		 * be allocated with our word list pointing into it.
		 */
		if (bufferLeft < len)
		{
			buffer = malloc(BUFSIZE);

			if (buffer == 0)
				fileError("Malloc failed");

			bufferLeft = BUFSIZE;
		}

		/*
		 * If the word list array needs resizing to hold more
		 * pointers for the words then do it.
		 */
		if (wordListSize == wordListAlloc)
		{
			wordListAlloc += LISTALLOC;

			wordList = (char **) realloc((char *) wordList,
				wordListAlloc * sizeof(char *));

			if (wordList == 0)
				fileError("Realloc failed");
		}

		/*
		 * Append a pointer to the word to the word list then
		 * copy the word into the buffer and advance past it.
		 */
		wordList[wordListSize++] = buffer;

		memcpy(buffer, line, len);

		buffer += len;
		bufferLeft -= len;
	}

	fclose(fp);

	if (wordListSize < MINWORDCOUNT)
		fileError("Not enough words found");
}


/*
 * Select a word from the word list at random.
 * The word list has already been read and only contains valid words.
 */
static const char *
pickWord(void)
{
	int index = randomRange(wordListSize);

	return wordList[index];
}


/*
 * Obtain a pseudo-random number.
 * The returned number is in the interval [0, range - 1].
 * This always calculates the random numbers modulo an odd number
 * (while ignoring the occasional bad value if we are unlucky).
 * This gives good results even for a range which is a power of two.
 */
static int
randomRange(int range)
{
	int answer;

	if (range == 0)
		return 0;

	do
	{
		answer = (int) (random() % (range | 1));
	}
	while (answer == range);

	return answer;
}


/*
 * Print the usage and quit.
 */
static void
usage(void)
{
	fprintf(stderr, "Usage: hangman [-c] [dictFile]\n");

	exit(1);
}


/*
 * Error routine for file reading.
 */
static void
fileError(const char * str)
{
	fprintf(stderr, "Error reading dictionary file %s: %s\n", wordFile, str);

	exit(1);
}

/* END CODE */
